use crate::goban::{Coordinate, Goban};
use config::define_config;
use config_derive::ConfigOption;
use serde::{Deserialize, Serialize};
use sgf::GameNode;
use std::{cell::RefCell, collections::VecDeque, fmt, path::PathBuf, time::Duration};
use thiserror::Error;
use uuid::Uuid;

define_config! {
    LibraryPath(LibraryPath),
    Me(Me),
}

#[derive(Clone, Debug, PartialEq, Serialize, Deserialize, ConfigOption)]
pub struct LibraryPath(pub PathBuf);

impl std::ops::Deref for LibraryPath {
    type Target = PathBuf;
    fn deref(&self) -> &Self::Target {
        &self.0
    }
}

impl From<String> for LibraryPath {
    fn from(s: String) -> Self {
        Self(PathBuf::from(s))
    }
}

#[derive(Clone, Debug, PartialEq, Serialize, Deserialize, ConfigOption)]
pub struct Me(Player);

impl std::ops::Deref for Me {
    type Target = Player;
    fn deref(&self) -> &Self::Target {
        &self.0
    }
}

#[derive(Debug, PartialEq, Error)]
pub enum BoardError {
    #[error("Position is invalid")]
    InvalidPosition,
    #[error("Self-capture is forbidden")]
    SelfCapture,
    #[error("Ko")]
    Ko,
}

#[derive(Clone, Copy, Debug, PartialEq, Hash, Eq, Serialize, Deserialize)]
pub enum Color {
    Black,
    White,
}

impl From<&sgf::Color> for Color {
    fn from(c: &sgf::Color) -> Self {
        match c {
            sgf::Color::Black => Self::Black,
            sgf::Color::White => Self::White,
        }
    }
}

impl fmt::Display for Color {
    fn fmt(&self, formatter: &mut fmt::Formatter<'_>) -> Result<(), fmt::Error> {
        match self {
            Color::Black => write!(formatter, "Black"),
            Color::White => write!(formatter, "White"),
        }
    }
}

#[derive(Clone, Copy, Debug, PartialEq, Eq, Serialize, Deserialize)]
pub struct Size {
    pub width: u8,
    pub height: u8,
}

impl Default for Size {
    fn default() -> Self {
        Self {
            width: 19,
            height: 19,
        }
    }
}

/// AppState stores all of the important state to a full running version of the application.
/// However, this version of AppState is in pretty sorry shape.
///
/// What are the states of the app?
///
/// - in review
/// - in a game
/// - connections to the internet
/// - connections to local applications, such as Leela Zero
/// - the current configuration
/// - the games database
/// - If in a game, the current state of the game. Delegated to GameState.
/// - If in review, the current state of the review.
///
/// Some of these states are concurrent. It's quite possible to have online connections running and
/// to be reviewing a game while, for instance, waiting for an opponent on OGS to make a move.
///
/// I get to ignore a lot of these things for now. Not playing online. Not playing at all,
/// actually. We'll come back to that.
///
/// Plus, it gets more fuzzy, because some of the application state is really UI state. For
/// instance, the state of a game review is purely UI.
///
/// So... AppState probably isn't great for now, but maybe it will become so later. I think I'm
/// going to ignore it until I need it.
/*
#[derive(Debug)]
pub struct AppState {
    pub game: Option<GameState>,
    pub database: Database,
}

impl AppState {
    pub fn new(database_path: LibraryPath) -> Self {
        Self {
            game: Some(GameState::default()),
            database: Database::open_path(database_path.to_path_buf()).unwrap(),
        }
    }

    /*
    pub fn place_stone(&mut self, req: PlayStoneRequest) {
        if let Some(ref mut game) = self.game {
            let _ = game.place_stone(Coordinate {
                column: req.column,
                row: req.row,
            });
        }
    }
    */
}
*/

#[derive(Clone, Copy, Debug, PartialEq, Eq, Serialize, Deserialize)]
pub enum Rank {
    Kyu(u8),
    Dan(u8),
    Pro(u8),
}

impl TryFrom<&str> for Rank {
    type Error = String;

    fn try_from(_: &str) -> Result<Rank, Self::Error> {
        Ok(Rank::Kyu(15))
    }
}

impl From<Rank> for String {
    fn from(r: Rank) -> String {
        match r {
            Rank::Kyu(v) => format!("{} kyu", v),
            Rank::Dan(v) => format!("{} dan", v),
            Rank::Pro(v) => format!("{} pro", v),
        }
    }
}

#[derive(Clone, Debug, PartialEq, Serialize, Deserialize)]
pub struct Player {
    pub name: String,
    pub rank: Option<Rank>,
}

#[derive(Debug)]
pub struct GameState {
    pub board: Goban,
    pub past_positions: Vec<Goban>,

    pub conversation: Vec<String>,
    pub current_player: Color,

    pub white_player: Player,
    pub black_player: Player,

    pub white_clock: Duration,
    pub black_clock: Duration,
}

impl Default for GameState {
    fn default() -> Self {
        Self {
            board: Goban::new(),
            past_positions: vec![],
            conversation: vec![],
            current_player: Color::Black,
            white_player: Player {
                name: "".to_owned(),
                rank: None,
            },
            black_player: Player {
                name: "".to_owned(),
                rank: None,
            },
            white_clock: Duration::from_secs(600),
            black_clock: Duration::from_secs(600),
        }
    }
}

impl GameState {
    // Legacy code. I recall that this is no longer used (but will be used again) because I
    // commented out so much code when I was overhauling the architecture of this app.
    #[allow(dead_code)]
    fn place_stone(&mut self, coordinate: Coordinate) -> Result<(), BoardError> {
        let board = self.board.clone();
        let new_board = board.place_stone(coordinate, self.current_player)?;

        if self.past_positions.contains(&new_board) {
            return Err(BoardError::Ko);
        }

        self.past_positions.push(self.board.clone());
        self.board = new_board;
        match self.current_player {
            Color::White => self.current_player = Color::Black,
            Color::Black => self.current_player = Color::White,
        };
        Ok(())
    }
}

// To properly generate a tree, I need to know how deep to go. Then I can backtrace. Each node
// needs to have a depth. Given a tree, the depth of the node is just the distance from the root.
// This seems obvious, but I had to write it to discover how important that fact was.
//
// So, what is the maximum depth of the tree? Follow all paths and see how far I get in every case.
// I could do this by just generating an intermediate tree and numbering each level.
pub struct Tree<T> {
    nodes: Vec<Node<T>>,
}

#[derive(Debug)]
pub struct Node<T> {
    pub id: usize,
    node: T,
    parent: Option<usize>,
    depth: usize,
    width: RefCell<Option<usize>>,
    children: Vec<usize>,
}

impl<T> Tree<T> {
    fn new(root: T) -> Self {
        Tree {
            nodes: vec![Node {
                id: 0,
                node: root,
                parent: None,
                depth: 0,
                width: RefCell::new(None),
                children: vec![],
            }],
        }
    }

    pub fn node(&self, idx: usize) -> &T {
        &self.nodes[idx].node
    }

    // Add a node to the parent specified by parent_idx. Return the new index. This cannot be used
    // to add the root node, but the constructor should handle that, anyway.
    fn add_node(&mut self, parent_idx: usize, node: T) -> usize {
        let next_idx = self.nodes.len();
        let parent = &self.nodes[parent_idx];

        self.nodes.push(Node {
            id: next_idx,
            node,
            parent: Some(parent_idx),
            depth: parent.depth + 1,
            width: RefCell::new(None),
            children: vec![],
        });

        let parent = &mut self.nodes[parent_idx];
        parent.children.push(next_idx);
        next_idx
    }

    pub fn max_depth(&self) -> usize {
        self.nodes.iter().fold(
            0,
            |max, node| if node.depth > max { node.depth } else { max },
        )
    }

    // Since I know the width of a node, now I want to figure out its placement in the larger
    // scheme of things.
    //
    // One thought I have is that I could just develop a grid virtually and start placing nodes.
    // Whenever I notice a collision, I can just move the node over. But I'd like to see if I can
    // be a bit smarter than doing it as just a vec into which I place things, as though it's a
    // game board. So, given a game node, I want to figure out it's position along the X axis.
    //
    // Just having the node is greatly insufficient. I can get better results if I'm calculating
    // the position of its children.
    //
    // indent represents the indentation that should be applied to all children in this tree. It
    // amounts to the position of the parent node.
    //
    // When drawing nodes, I don't know how to persist the level of indent.
    pub fn position(&self, idx: usize) -> (usize, usize) {
        let node = &self.nodes[idx];
        match node.parent {
            Some(parent_idx) => {
                let (_parent_row, parent_column) = self.position(parent_idx);
                let parent = &self.nodes[parent_idx];
                let sibling_width = parent
                    .children
                    .iter()
                    .take_while(|n| **n != node.id)
                    .fold(0, |acc, n| acc + self.width(*n));
                println!("[{}] sibling width {}", idx, sibling_width);
                (node.depth, parent_column + sibling_width)
            }

            // Root nodes won't have a parent, so just put them in the first column
            None => (0, 0),
        }
    }

    // Given a node, do a postorder traversal to figure out the width of the node based on all of
    // its children. This is equivalent to the widest of all of its children at all depths.
    //
    // There are some collapse rules that I could take into account here, but that I haven't
    // figured out yet. If two nodes are side by side, and one of them has some wide children but
    // the other has no children, then they are effectively the same width. The second node only
    // needs to be moved out if it has children that would overlap the children of the first node.
    //
    // My algorithm right now is likely to generate unnecessarily wide trees in a complex game
    // review.
    fn width(&self, id: usize) -> usize {
        println!("[{}] calculating width", id);
        let node = &self.nodes[id];
        if let Some(width) = *node.width.borrow() {
            return width;
        }

        let width = node
            .children
            .iter()
            .fold(0, |acc, child| acc + self.width(*child));
        let width = if width == 0 { 1 } else { width };
        println!("[{}] width: {}", id, width);
        *node.width.borrow_mut() = Some(width);

        width
    }

    pub fn bfs_iter<'a>(&'a self) -> BFSIter<T> {
        let mut queue = VecDeque::new();
        queue.push_back(&self.nodes[0]);
        BFSIter { tree: self, queue }
    }
}

impl<'a> From<&'a GameNode> for Tree<Uuid> {
    fn from(root: &'a GameNode) -> Self {
        fn add_subtree<'a>(tree: &mut Tree<Uuid>, parent_idx: usize, node: &'a GameNode) {
            let idx = tree.add_node(parent_idx, node.id());

            let children = match node {
                GameNode::MoveNode(node) => &node.children,
                GameNode::SetupNode(node) => &node.children,
            };

            for child in children {
                add_subtree(tree, idx, child);
            }
        }

        let mut tree = Tree::new(root.id());

        let children = match root {
            GameNode::MoveNode(node) => &node.children,
            GameNode::SetupNode(node) => &node.children,
        };

        for node in children {
            add_subtree(&mut tree, 0, node);
        }

        tree
    }
}

pub struct BFSIter<'a, T> {
    tree: &'a Tree<T>,
    queue: VecDeque<&'a Node<T>>,
}

impl<'a, T> Iterator for BFSIter<'a, T> {
    type Item = &'a Node<T>;

    fn next(&mut self) -> Option<Self::Item> {
        let retval = self.queue.pop_front();
        if let Some(ref retval) = retval {
            retval
                .children
                .iter()
                .for_each(|idx| self.queue.push_back(&self.tree.nodes[*idx]));
        }
        retval
    }
}

#[cfg(test)]
mod test {
    use super::*;
    use cool_asserts::assert_matches;
    use sgf::{Move, MoveNode};

    #[test]
    fn current_player_changes_after_move() {
        let mut state = GameState::default();
        assert_eq!(state.current_player, Color::Black);
        state.place_stone(Coordinate { column: 9, row: 9 }).unwrap();
        assert_eq!(state.current_player, Color::White);
    }

    #[test]
    fn current_player_remains_the_same_after_self_capture() {
        let mut state = GameState::default();
        state.board = Goban::from_coordinates(
            vec![
                (Coordinate { column: 17, row: 0 }, Color::White),
                (Coordinate { column: 17, row: 1 }, Color::White),
                (Coordinate { column: 18, row: 1 }, Color::White),
            ]
            .into_iter(),
        )
        .unwrap();
        state.current_player = Color::Black;

        assert_eq!(
            state.place_stone(Coordinate { column: 18, row: 0 }),
            Err(BoardError::SelfCapture)
        );
        assert_eq!(state.current_player, Color::Black);
    }

    #[test]
    fn ko_rules_are_enforced() {
        let mut state = GameState::default();
        state.board = Goban::from_coordinates(
            vec![
                (Coordinate { column: 7, row: 9 }, Color::White),
                (Coordinate { column: 8, row: 8 }, Color::White),
                (Coordinate { column: 8, row: 10 }, Color::White),
                (Coordinate { column: 9, row: 9 }, Color::White),
                (Coordinate { column: 10, row: 9 }, Color::Black),
                (Coordinate { column: 9, row: 8 }, Color::Black),
                (Coordinate { column: 9, row: 10 }, Color::Black),
            ]
            .into_iter(),
        )
        .unwrap();

        state.place_stone(Coordinate { column: 8, row: 9 }).unwrap();
        assert_eq!(
            state.place_stone(Coordinate { column: 9, row: 9 }),
            Err(BoardError::Ko)
        );
    }

    // A
    // B G   H
    // C     I
    // D E F
    #[test]
    fn it_can_calculate_depth_from_game_tree() {
        let mut node_a = MoveNode::new(sgf::Color::Black, Move::Move("dp".to_owned()));
        let mut node_b = MoveNode::new(sgf::Color::Black, Move::Move("dp".to_owned()));
        let mut node_c = MoveNode::new(sgf::Color::Black, Move::Move("dp".to_owned()));
        let node_d = MoveNode::new(sgf::Color::Black, Move::Move("dp".to_owned()));
        let node_e = MoveNode::new(sgf::Color::Black, Move::Move("dp".to_owned()));
        let node_f = MoveNode::new(sgf::Color::Black, Move::Move("dp".to_owned()));
        let node_g = MoveNode::new(sgf::Color::Black, Move::Move("dp".to_owned()));
        let mut node_h = MoveNode::new(sgf::Color::Black, Move::Move("dp".to_owned()));
        let node_i = MoveNode::new(sgf::Color::Black, Move::Move("dp".to_owned()));

        node_c.children.push(GameNode::MoveNode(node_d));
        node_c.children.push(GameNode::MoveNode(node_e));
        node_c.children.push(GameNode::MoveNode(node_f));

        node_b.children.push(GameNode::MoveNode(node_c));

        node_h.children.push(GameNode::MoveNode(node_i));

        node_a.children.push(GameNode::MoveNode(node_b));
        node_a.children.push(GameNode::MoveNode(node_g));
        node_a.children.push(GameNode::MoveNode(node_h));

        let game_tree = GameNode::MoveNode(node_a);

        let tree = Tree::from(&game_tree);

        assert_eq!(tree.max_depth(), 3);
    }

    // A
    // B G   H
    // C     I
    // D E F
    #[test]
    fn it_calculates_horizontal_position_of_nodes() {
        let mut node_a = MoveNode::new(sgf::Color::Black, Move::Move("dp".to_owned()));
        let mut node_b = MoveNode::new(sgf::Color::Black, Move::Move("dp".to_owned()));
        let mut node_c = MoveNode::new(sgf::Color::Black, Move::Move("dp".to_owned()));
        let node_d = MoveNode::new(sgf::Color::Black, Move::Move("dp".to_owned()));
        let node_e = MoveNode::new(sgf::Color::Black, Move::Move("dp".to_owned()));
        let node_f = MoveNode::new(sgf::Color::Black, Move::Move("dp".to_owned()));
        let node_g = MoveNode::new(sgf::Color::Black, Move::Move("dp".to_owned()));
        let mut node_h = MoveNode::new(sgf::Color::Black, Move::Move("dp".to_owned()));
        let node_i = MoveNode::new(sgf::Color::Black, Move::Move("dp".to_owned()));

        node_c.children.push(GameNode::MoveNode(node_d));
        node_c.children.push(GameNode::MoveNode(node_e));
        node_c.children.push(GameNode::MoveNode(node_f));

        node_b.children.push(GameNode::MoveNode(node_c));

        node_h.children.push(GameNode::MoveNode(node_i));

        node_a.children.push(GameNode::MoveNode(node_b));
        node_a.children.push(GameNode::MoveNode(node_g));
        node_a.children.push(GameNode::MoveNode(node_h));

        let game_tree = GameNode::MoveNode(node_a);

        let tree = Tree::from(&game_tree);

        assert_eq!(tree.position(2), (2, 0));
        assert_eq!(tree.position(1), (1, 0));
        assert_eq!(tree.position(0), (0, 0));
        assert_eq!(tree.position(4), (3, 1));
        assert_eq!(tree.position(5), (3, 2));
        assert_eq!(tree.position(6), (1, 3));
        assert_eq!(tree.position(7), (1, 4));
    }

    #[test]
    fn breadth_first_iter() {
        let mut node_a = MoveNode::new(sgf::Color::Black, Move::Move("dp".to_owned()));
        let mut node_b = MoveNode::new(sgf::Color::Black, Move::Move("dp".to_owned()));
        let mut node_c = MoveNode::new(sgf::Color::Black, Move::Move("dp".to_owned()));
        let node_d = MoveNode::new(sgf::Color::Black, Move::Move("dp".to_owned()));
        let node_e = MoveNode::new(sgf::Color::Black, Move::Move("dp".to_owned()));
        let node_f = MoveNode::new(sgf::Color::Black, Move::Move("dp".to_owned()));
        let node_g = MoveNode::new(sgf::Color::Black, Move::Move("dp".to_owned()));
        let mut node_h = MoveNode::new(sgf::Color::Black, Move::Move("dp".to_owned()));
        let node_i = MoveNode::new(sgf::Color::Black, Move::Move("dp".to_owned()));

        node_c.children.push(GameNode::MoveNode(node_d.clone()));
        node_c.children.push(GameNode::MoveNode(node_e.clone()));
        node_c.children.push(GameNode::MoveNode(node_f.clone()));

        node_b.children.push(GameNode::MoveNode(node_c.clone()));

        node_h.children.push(GameNode::MoveNode(node_i.clone()));

        node_a.children.push(GameNode::MoveNode(node_b.clone()));
        node_a.children.push(GameNode::MoveNode(node_g.clone()));
        node_a.children.push(GameNode::MoveNode(node_h.clone()));

        let game_tree = GameNode::MoveNode(node_a.clone());

        let tree = Tree::from(&game_tree);

        let mut iter = tree.bfs_iter();

        assert_matches!(iter.next(), Some(Node { node: uuid, .. }) => assert_eq!(*uuid, node_a.id));
        assert_matches!(iter.next(), Some(Node { node: uuid, .. }) => assert_eq!(*uuid, node_b.id));
        assert_matches!(iter.next(), Some(Node { node: uuid, .. }) => assert_eq!(*uuid, node_g.id));
        assert_matches!(iter.next(), Some(Node { node: uuid, .. }) => assert_eq!(*uuid, node_h.id));
        assert_matches!(iter.next(), Some(Node { node: uuid, .. }) => assert_eq!(*uuid, node_c.id));
        assert_matches!(iter.next(), Some(Node { node: uuid, .. }) => assert_eq!(*uuid, node_i.id));
        assert_matches!(iter.next(), Some(Node { node: uuid, .. }) => assert_eq!(*uuid, node_d.id));
        assert_matches!(iter.next(), Some(Node { node: uuid, .. }) => assert_eq!(*uuid, node_e.id));
        assert_matches!(iter.next(), Some(Node { node: uuid, .. }) => assert_eq!(*uuid, node_f.id));
    }
}
